package columnGeneration;

import java.util.ArrayList;
import java.util.List;

import ilog.concert.IloException;
import instance.Instance;
import instance.Parameters;
import instance.Task;
import pricingProblem.PricingProblemSolver;
import util.Pair;

/**
 * Simple column generation framework to solve the LP-relaxation of a crew
 * scheduling problem.
 * 
 * @author B.T.C. van Rossum
 */
public class ColumnGeneration
{
	/**
	 * Solve the LP-relaxation of an instance of the crew scheduling problem using a
	 * specified pricing problem solver, and add the time spent pricing and number
	 * of pricing iterations to an array.
	 * 
	 * @param instance   instance of the crew scheduling problem
	 * @param solver     pricing problem solver
	 * @param statistics array that stores cumulative time spent pricing and number
	 *                   of pricing iterations
	 * @throws IloException exception thrown by CPLEX
	 */
	public static void applyColumnGeneration(Instance instance, PricingProblemSolver solver, double[] statistics)
			throws IloException
	{
		// Initialise RMP.
		RMP rmp = new RMP(instance);
		DisjointSelector selector = new DisjointSelector(Parameters.SIMILARITY_THRESHOLD);

		// Add slack duties.
		addSlackDuties(rmp, instance);

		// Main loop.
		while (true)
		{
			// Solve RMP.
			rmp.solve();

			// Solve pricing problems.
			statistics[1]++;
			statistics[0] -= System.currentTimeMillis();
			solver.prepareGraphs(rmp);
			List<Pair<Duty, Double>> duties = solver.generateDuties();
			statistics[0] += System.currentTimeMillis();

			// Add duties to RMP or terminate if no duties exist.
			if (duties.size() > 0)
			{
				// Add a subset of the generated columns to the RMP.
				for (Duty duty : selector.selectColumns(duties))
				{
					rmp.addDuty(duty);
				}
			}
			else
			{
				// No duties are found, the RMP is solved to optimality.
				break;
			}
		}

		// Clean model.
		rmp.clean();
	}

	/**
	 * Add a slack variable for each task in the instance such that a feasible
	 * solution always exists.
	 * 
	 * @param rmp      restricted master problem
	 * @param instance instance of the crew scheduling problem
	 * @throws IloException exception thrown by CPLEX
	 */
	private static void addSlackDuties(RMP rmp, Instance instance) throws IloException
	{
		// The cost of a slack variable is the maximum cost of a feasible duty.
		int slackCost = Parameters.FIXED_DUTY_COST + Parameters.MAX_DUTY_LENGTH * Parameters.VARIABLE_DUTY_COST;
		for (Task task : instance.getTasks())
		{
			List<Task> tasks = new ArrayList<>();
			tasks.add(task);
			rmp.addDuty(new Duty(slackCost, tasks));
		}
	}
}
